/*
 * Copyright (c) 1999, 2000, 2001, 2002 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#if (FREEBSD_4 || FREEBSD_5)
#include <sys/soundcard.h>
#endif	/* FREEBSD_4 */

#ifdef	NETBSD
#include <sys/ioccom.h>
#include <soundcard.h>
#endif	/* NETBSD */

#ifdef	LINUX
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#endif	/* LINUX */

#include <libdv/dv.h>

#include "xdvshow-audio.h"

/** \brief is the audio device file descriptor.
 */

int audio_fd;

static int _init_audio_device __P((dv_audio_t *));

/** \brief Public interface for _init_audio_device fuction.
 *
 * xdvshow_init_audio opens given audio device and then calls internal
 * function _init_audio_device.
 *
 * \param *audio_dev name of the audio device
 * \param *audio dv_audio_t structure
 *
 * \return 1 for success. -1 otherwise. The function also sets errno when anything goes wrong.
 */

int
xdvshow_init_audio(char *audio_dev, dv_audio_t *audio)
{
  int ret;

  if (audio_dev == NULL) {
    return(-1);
  }

  audio_fd = open(audio_dev, O_WRONLY);
  if (audio_fd < 0) {
    perror("open audio");
    return(-1);
  }

  ret = _init_audio_device(audio);
  if (ret < 0) {
    return(-1);
  }

  return(1);
}

/** \brief Audio playing function.
 *
 * xdvshow_play_audio() just writes the previously decoded audio 
 * data to the audio device. The device must be initialized in advance
 * by xdvshow_init_audio().
 *
 * \param *audio is a dv_audio_t structure
 * \param **audio_buffers buffers storing decoded audio data
 *
 * \return 1. There is no other return in this function.
 */

int
xdvshow_play_audio(dv_audio_t *audio, int16_t **audio_buffers)
{
  /*
  int ret;
  */
  int ch, i, j=0;
  static int16_t outbuf[DV_AUDIO_MAX_SAMPLES * 4];

  for (i=0; i<audio->samples_this_frame; i++) {
    for (ch=0; ch<audio->num_channels; ch++) {
      outbuf[j] = audio_buffers[ch][i];
      j++;
    }
  }

  write(audio_fd,
        outbuf,
        audio->samples_this_frame * audio->num_channels * sizeof(u_int16_t));

  /*
  ret = ioctl(audio_fd, SNDCTL_DSP_POST, NULL);
  if (ret < 0) {
    perror("SNDCTL_DSP_POST");
    return (-1);
  }
  */

  return(1);
}

/** \brief This is the internal audio initialization fuction.
 *
 *  Function has its own parts for Linux and FreeBSD. Perhaps the
 *  initialization code should be the same for both systems. 
 *  Anyway it's necessary to set the endianity, number of channels 
 *  and sampling rate. These values must be set strictly in this 
 *  order to intialize the audio properly.
 *
 * \param *audio is a dv_audio_t structure
 *
 * \return 1 for success. -1 otherwise. The function also sets errno when anything goes wrong.
 */

static int
_init_audio_device(dv_audio_t *audio)
{
  int ret;
  int format;
  int fragment_request;
  int channels;
  int rate;

  /* 
   AFMT_S16_LE = little endian, 
   AFMT_S16_BE = big endian, 
   AFMT_S16_NE = native endianity of the machine;
  */
  format = AFMT_S16_NE;
  fragment_request = 0x000300B;
  channels = audio->num_channels;
  rate = audio->frequency;

#ifdef LINUX
  ret = ioctl(audio_fd, SNDCTL_DSP_SETFMT, &format);
  if (ret < 0) {
    perror("SNDCTL_DSP_SETFMT");
    return(-1);
  }

  ret = ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &channels);
  if (ret < 0) {
    perror("SNDCTL_DSP_CHANNELS");
    return (-1);
  }

  ret = ioctl(audio_fd, SNDCTL_DSP_SPEED, &rate);
  if (ret < 0) {
    perror("SNDCTL_DSP_SPEED");
    return (-1);
  }

#else /* FreeBSD and perhaps others */
  
  ret = ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &fragment_request);
  if (ret < 0) {
    perror("SNDCTL_DSP_SETFRAGMENT");
    return(-1);
  }

  ret = ioctl(audio_fd, SOUND_PCM_WRITE_BITS, &format);
  if (ret < 0) {
    perror("SOUND_PCM_WRITE_BITS");
    return(-1);
  }

  ret = ioctl(audio_fd, SOUND_PCM_WRITE_CHANNELS, &channels);
  if (ret < 0) {
    perror("SOUND_PCM_WRITE_CHANNELS");
    return(-1);
  }

  ret = ioctl(audio_fd, SOUND_PCM_WRITE_RATE, &rate);
  if (ret < 0) {
    perror("SOUND_PCM_WRITE_RATE");
    fprintf(stderr, "rate : %d\n", rate);
    return(-1);
  }
 
#endif

  return(1);
}
