#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <pthread.h>
#include <assert.h>

#include <libdv/dv_types.h>
#include <semaphore.h>
#include <errno.h>

#include "xdvshow-shm.h"
#include "xdvshow-flags.h"
#include "xdvshow-file.h"
#include "xdvshow-const.h"
#include "xdvshow-defs.h"

static int _xdvshow_read_file __P((void));

/** \brief File descriptor for reading from input file. 
  */
int _file;

/** \brief Public function for handling the input file. 
  * 
  * \param data is input filename (char *)
  * \return NULL on success, -1 on failure
  */
void *
xdvshow_prepare_file(void *data)
{
  char *filename = data;
  int ret;

  ret = _xdvshow_alloc_shm();
  if (ret < 0) {
    return((void *)1);
  }

  ret = _xdvshow_attach_shm();
  assert(ret==1);

  if (filename == NULL) {
    return((void *)-1);
  }

  _file = open(filename, O_RDONLY);
  if (_file == -1) {
    perror("open");

    if (kill(getpid(), SIGINT) != 0) {
	    perror("killing failed");
	    return((void *)-1);
    }
    
    return((void *)-1);
  }

  /* THIS IS SUPPOSED TO COME BACK AT THE END OF FILE */
  if(_xdvshow_read_file() != 0) {
      xdvshow_close_file();
	  return((void *)-1);
  }

  xdvshow_close_file();

  if (kill(getpid(), SIGINT) != 0) {
	  perror("killing failed");
	  return((void *)-1);
  }

  return NULL;
}

/** \brief Public function for closing input file 
  * 
  * The function closes input file usinf _file file descriptor. The function
  * has void type and it is used in xdvshow-main.c for registering function
  * pointer.
  */
void
xdvshow_close_file()
{
  if(close(_file) != 0) {
	  perror("closing file failed");
  }
}

/** \brief Internal function for reading and parsing the input file.
  * 
  * \return 0 on success, -1 on failure
  */
static int
_xdvshow_read_file()
{
  ssize_t n;

  u_char *dvdata;
  u_char *audio_dvdata;

#ifdef DEBUG
  int sem_value = 0;
  int frame_length = 0;
  int iDebugSemVal;
#endif /* DEBUG */

  u_int32_t dvdif[80 / 4];
  unsigned int sct, dseq, dbn;

  unsigned int dseq_max;
  unsigned int dseq_stream_max = 0;
  unsigned int dseq_stream_max_count = 0;

  switch (dv_format_type) {
    case e_dv_system_525_60:
      dseq_max = DSEQ_MAX_NTSC;
      break;

    case e_dv_system_625_50:
      dseq_max = DSEQ_MAX_PAL;
      break;

    default:
      return(-1);
  }

  dvdata = _xdvshow_shm.shm_frame->frame_buf->data;
  audio_dvdata = _xdvshow_shm.audio_shm_frame->frame_buf->data;

  while (read_running) {

    n = read(_file, (char *)&dvdif, 80);
    if (n < 1) {
      if (n == 0) { /*eof*/
        return(0);
      }
      perror("read");
      return(-1);
    }
    
#ifdef DEBUG
    frame_length += n;
#endif /* DEBUG */

    DPRINT_BARE(".");

    dbn  = (ntohl(dvdif[0]) >> 8) & 0xff;
    dseq = (ntohl(dvdif[0]) >> 20) & 0xf;
    sct  = (ntohl(dvdif[0]) >> 29) & 0x7;

    if (sct > 4) {
      DPRINT_BARE("\n");
      DPRINT("Data error: SCT > 4! sct = %d\n", sct);
      continue;
    }

    if (dseq > dseq_max) {
                DPRINT_BARE("\n");
                DPRINT("Data error: DSEQ > DSEQ_MAX! dseq = %d, dseq_max = %d\n", dseq, dseq_max);
          }
          if (dseq > DSEQ_MAX_PAL) {
                DPRINT_BARE("\n");
                DPRINT("Data error: DSEQ > DSEQ_MAX_PAL! dseq = %d, DSEQ_MAX_PAL = %d\n", dseq, DSEQ_MAX_PAL);
                continue;
          }

    if (dseq_stream_max_count >= 255) {
      if (dseq_stream_max != dseq_max &&(dseq_stream_max == DSEQ_MAX_PAL || dseq_stream_max == DSEQ_MAX_NTSC)) {
	dseq_max = dseq_stream_max;
      }
      dseq_stream_max_count = 0;
      dseq_stream_max = 0;
    } else {
      if (dseq_stream_max < dseq) {
	dseq_stream_max = dseq;
      }
      ++dseq_stream_max_count;
    }

    switch (sct) {
      case SCT_HEADER:
	if ((dseq * 150 * 80) <= (DVFRAME_MAX_SIZE - 80)) {
	  memcpy(&dvdata[dseq * 150 * 80], &dvdif, 80);
	} else {
	  DPRINT_BARE("\n");
	  DPRINT("memcpy droped - dvdata size %d exceeded by %d\n dseq = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 * 80) - DVFRAME_MAX_SIZE + 80), dbn);
	}
	break;
	
      case SCT_SUBCODE:
	if (dbn > 1) {
	  DPRINT_BARE("\n");
	  DPRINT("Data error: DBN > 1\ndbn = %d\n", dbn);
	  break;
	}
	if((dseq * 150 + 1 + dbn) *80 <= (DVFRAME_MAX_SIZE - 80)) {
	  memcpy(&dvdata[(dseq * 150 + 1 + dbn) * 80], &dvdif, 80);
	} else {
	  DPRINT_BARE("\n");
	  DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 1 + dbn)* 80 - DVFRAME_MAX_SIZE + 80), dbn);
	}	  
        break;
	
      case SCT_AAUX:
	break;

      case SCT_VAUX:
	if (dbn > 2) {
	  DPRINT_BARE("\n");
	  DPRINT("Data error: DBN > 2\ndbn = %d\n", dbn);
	}
	if((dseq * 150 + dbn + 3)*80 <= (DVFRAME_MAX_SIZE - 80)) {
          memcpy(&dvdata[(dseq * 150 + dbn + 3) * 80], &dvdif, 80);
	} else {
	  DPRINT_BARE("\n");
	  DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d, dbn = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 6 + dbn * 16)*80 - DVFRAME_MAX_SIZE + 80), dseq, dbn);
	}
        break;

      case SCT_AUDIO:
	if (dbn > 8) {
	  DPRINT_BARE("\n");
	  DPRINT("Data error: DBN > 8\ndbn = %d\n", dbn);
	  break;
	}
	DPRINT_BARE("*");
	if ((dseq * 150 + 6 + dbn * 16)*80 <= (DVFRAME_MAX_SIZE - 80)) {
          memcpy(&dvdata[(dseq * 150 + 6 + dbn*16) * 80], &dvdif, 80);
	} else {
	  DPRINT_BARE("\n");
	  DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d, dbn = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 6 + dbn * 16)*80 - DVFRAME_MAX_SIZE + 80), dseq, dbn);
	  break;
	}

	if (flags_use_audio) {
	  DPRINT_BARE("(*)");
	  if((dseq * 150 + 6 + dbn * 16) * 80  <= (DVFRAME_MAX_SIZE - 80)) {
  	    memcpy(&audio_dvdata[(dseq * 150 + 6 + dbn * 16) * 80], &dvdif, 80);
	  } else {
	    DPRINT_BARE("\n");
	    DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d, dbn = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 6 + dbn * 16)*80 - DVFRAME_MAX_SIZE + 80), dseq, dbn);
	    break;
	  }

	  if (dseq == dseq_max && dbn == 8) {

#ifdef DEBUG
	    if (sem_getvalue(&audio_ready, &iDebugSemVal)) {
	      perror("Unable to get value of audio_ready semaphore!");
	      exit(EXIT_FAILURE);
	    }

	    DPRINT("semaphore audio_ready value: %d\n", iDebugSemVal);

	    if (sem_getvalue(&audio_empty, &iDebugSemVal)) {
	      perror("Unable to get value of audio_empty semaphore!");
	      exit(EXIT_FAILURE);
	    }

	    DPRINT("semaphore audio_empty value: %d\n", iDebugSemVal);
#endif /* DEBUG */

	    if ( !( ((flags_format_normal == 0) && (flags_format_wide == 0))
		  || ((flags_format_normal != 0) && (flags_format_wide != 0)) ) )
	    	{
		  sem_post(&audio_ready);
		  ts_sem_wait(&audio_empty);
		}

	    _xdvshow_shm.audio_shm_frame = _xdvshow_shm.audio_shm_frame->next;
	    audio_dvdata = _xdvshow_shm.audio_shm_frame->frame_buf->data;
	  }
	}
	  
	
        break;

      case SCT_VIDEO:
        memcpy(&dvdata[(dseq * 150 + 7 + dbn + dbn/15) * 80], &dvdif, 80);
        if (dseq == dseq_max && dbn == 134) {
            _xdvshow_shm.shm_frame->frame_buf->lock = DVFRAME_DATA_READY;

#ifdef DEBUG
	    DPRINT("RECEIVED frame length: %d.\n", frame_length);
	    frame_length = 0;
	    
	    DPRINT("RECEIVED frame no. %d\n", sem_value++);

	    if(sem_getvalue(&video_ready, &iDebugSemVal)) {
		    perror("Unable to get value of video_ready semaphore!");
		    exit(EXIT_FAILURE);
	    }
	    DPRINT("semaphore video_ready value: %d\n", iDebugSemVal);

	    if(sem_getvalue(&video_empty, &iDebugSemVal)) {
		    perror("Unable to get value of video_empty semaphore!");
		    exit(EXIT_FAILURE);
	    }
	    DPRINT("semaphore video_empty value: %d\n", iDebugSemVal);
#endif /* DEBUG */

	    if (sem_post(&video_ready) < 0) {
		    perror("sem_post");
		    assert(0);
	    }
	    ts_sem_wait(&video_empty);

      	    _xdvshow_shm.shm_frame = _xdvshow_shm.shm_frame->next;
            dvdata = _xdvshow_shm.shm_frame->frame_buf->data;
        }
        break;

      default:
	fprintf(stderr, "SCT: %d\n", sct);
        fprintf(stderr, "INVALID FILE FORMAT!!!\n");
        return(-1);
    }
  }

  /* This should be unreachable but it must be verified. */
  assert(0);
  return(-1);
}
