/*
 * Copyright (c) 1999, 2000, 2001, 2002 WIDE Project
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/shm.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <pthread.h>

#ifdef	LINUX
#include <getopt.h>
#endif	/* LINUX */

#include <libdv/dv.h>

#include <xdvshow-info.h>

#if	(HAVE_SDL || HAVE_SDL11)
#include <xdvshow-sdl.h>
#endif	/* HAVE_SDL */

#include <xdvshow-x11.h>

#include "xdvshow-shm.h"
#include "xdvshow-rtp.h"
#include "xdvshow-file.h"
#include "xdvshow-ieee1394.h"
#include "xdvshow-audio.h"
#include "xdvshow-flags.h"
#include "xdvshow-const.h"
#include "xdvshow-defs.h"

int main_running = 1;
int video_running = 1;
int audio_running = 1;
int read_running = 1;
int ping_running = 1;

int flags_stereo = 0;
int flags_fulscreen = 0;
#if (HAVE_SDL || HAVE_SDL11)
int flags_sdl = 1;
int flags_x11 = 0;
#else
int flags_sdl = 0;
int flags_x11 = 1;
#endif
int flags_sdlnooverlay = 0;
int flags_file = 0;
int flags_zoom = 0;
int flags_scale = 0;
int flags_dump = 0;
int flags_deinterlace = 0;
int flags_reflector = 0;
int flags_mcast = 0;
int flags_use_audio = 0;

int flags_format_wide = 0;
int flags_format_normal = 0;

int flags_half_size = 0;
int flags_half_framerate = 0;

int window_x = -1;
int window_y = -1;
int window_w = 0;
int window_h = 0;
unsigned long window_id = 0;

u_int32_t	flags = 0x0;
u_int32_t	dv_format_type = e_dv_system_625_50;
int (*input_func) __P((u_char **));
int (*input_unlock_func) __P(());

int (*audio_input_func) __P((u_char **));
int (*audio_unlock_func) __P(());

char *prog_name = XDVSHOW_PROGNAME;

void *video_thread_func __P((void *));
void *audio_thread_func __P((void *));

u_char *videodata;

u_char *audioframe;
int16_t *audio_buffers[4];


struct xdvshow_rtp_param rtp_param;

int dump_fd;
char *reflector_address;

struct timeval currtime;
long frametime_usec;
long frametimeadd_usec = 27000;
long framedelay_usec;

dv_decoder_t *dv_dec;

pthread_t read_thread;
pthread_t ping_thread;
pthread_t video_thread;

/** \brief Function that processes SIGINT signal and tries to terminate
  * xdvshow.
  *
  * It waits for ping_thread, video_thread, and read_thread termination and
  * closes dump filehandle (if needed) and cleans up (detaches) shared memory.
  * It also frees up DV decoder.
  * \param signo Signal number
  * \return exits with EXIT_SUCCESS 
  */
static void 
sigint_signal(int signo) {

  void *thread_result;
  int i;
  int ret;

  assert(signo == SIGINT);

  DPRINT_BARE("\n");
  DPRINT("Entering SIGINT handling.\n");

  if (flags_reflector) {
    DPRINT("Waiting for ping thread to finish.\n");
    ret = pthread_join(ping_thread, &thread_result);
    if (ret < 0) {
      perror("ping_thread join failed");
    }
  }

  /* XXX - when calling pthread_cancel(video_thread) here, I get blocking
     behavior again. Perhaps this is due to buggy libc_r on FreeBSD 5.X
	 (X < 3) and will be corrected as a part of libpthead on 5.3. A backtrace
	 of the blocking looks usually like:
	 #0  0x281fdd1b in pthread_testcancel () from /usr/lib/libkse.so.1
     #1  0x281f649c in pthread_mutexattr_init () from /usr/lib/libkse.so.1
     #2  0x281fa524 in pthread_setconcurrency () from /usr/lib/libkse.so.1
     #3  0x281f11b6 in _nanosleep () from /usr/lib/libkse.so.1
     #4  0x2824420f in sleep () from /lib/libc.so.5
     #5  0x281e712f in sleep () from /usr/lib/libkse.so.1
     #6  0x0804b72b in audio_thread_func (data=0x8063000) at xdvshow-main.c:627
     #7  0x0804b337 in main (argc=3, argv=0xbfbfe40c) at xdvshow-main.c:509
     #8  0x08049822 in _start ()
     */
  DPRINT("Waiting for video thread to finish.\n");

  /* In some pathological cases (other init failed for some reason) 
   * the video thread may not be initialized. */
    if (video_thread) {
#ifndef FREEBSD_5
    pthread_cancel(video_thread);
#endif /* FREEBSD_5 */
    ret = pthread_join(video_thread, &thread_result);
      if (ret < 0) {
      perror("video_thread join failed");
    }
  } else {
    DPRINT("No video thread initialized.\n");
  }
    

  DPRINT("Waiting for read thread to finish.\n");
  /* XXX - pthread_cancel is blocking here on linux machine (2.4.22 kernel)
           when using SDL for video displaying.
   */
#ifdef LINUX26
  sleep(1);
  pthread_cancel(read_thread);
#endif /* LINUX26 */
  /* XXX - some other black magic here :( xdvshow terminates here with x11 display */
  pthread_kill(read_thread, SIGTERM);
  ret = pthread_join(read_thread, &thread_result);
  if (ret < 0) {
    perror("read_thread join failed");
  }

  if (flags_dump) {
    DPRINT("Closing dump file.\n");
    close(dump_fd);
  }

#if (HAVE_SDL || HAVE_SDL11)
  if (flags_sdl != 0) { 
    DPRINT("Closing SDL.\n");
    xdvshow_close_window();
  }
#endif /* HAVE_SDL */
  if (flags_x11 != 0) {
    DPRINT("Closing X11.\n");
    xdvshow_x11_close_window();
  }

  DPRINT("Freeing memory.\n");

  if (flags_use_audio) {
    for (i = 0; i < 4; i++) {
      free(audio_buffers[i]);
    }
  }
  
  dv_decoder_free(dv_dec);

  if(shmdt(_xdvshow_shm.shm_buf) != 0) {
	  perror("shmdt on _xdvshow_shm.shm_buf");
  }
  if (shmctl(_xdvshow_shm.shmid, IPC_RMID, NULL) != 0) {
	  perror("shmctl on _xdvshow_shm.shmid");
  }

  printf("Exit\n");
  exit (EXIT_SUCCESS);
}

/** \brief Main entry fucntion of the prgram.
  *
  * The function parses input parameters using getopt and spawn pthread threads
  * based on input parameters.
  * \param argc number of command line arguments
  * \param *argv[] array of command line arguments
  * \return EXIT_SUCCESS when successful, EXIT_FAILURE otherwise
  */
int
main(int argc, char *argv[])
{
  int ret;

  u_char *dvframe;

  int quality;

  int i;

  char *audio_dev = NULL;

  extern char *optarg;
  int op;
  char *opt_tmp = NULL;

  void *open_option = NULL;

  void *(*input_open) __P((void *));
  void (*input_close) __P((void));

  pthread_attr_t bigstack;
  
  if (signal(SIGINT, sigint_signal) == SIG_ERR) {
    fprintf(stderr, "Cannot install signal handler.\n");
  }

  memset(&rtp_param, 0, sizeof(rtp_param));

  dv_dec = dv_decoder_new(FALSE, FALSE, FALSE);
  if (dv_dec == NULL) {
    fprintf(stderr, "ERROR : dv_decoder_new\n");
    return(EXIT_FAILURE);
  }

  dv_dec->quality = DV_QUALITY_BEST;

  input_open  = xdvshow_prepare_rtp;
  input_func  = xdvshow_read_shm;
  input_unlock_func = xdvshow_read_unlock;
  audio_input_func = xdvshow_read_audio_shm;
  audio_unlock_func = xdvshow_read_audio_unlock;
  input_close = xdvshow_close_shm;
  open_option = &rtp_param;

  rtp_param.s_addr.ss_family = (sa_family_t) AF_INET;
  rtp_param.port = 12000;

  while ((op = getopt(argc, argv, "Q:A:vhf:46P:j:r:odDFzZSixsC:R:W:IH")) > 0) {
    switch (op) {
      case 'v':
        show_version();
        exit(EXIT_SUCCESS);

  case 'Q':
      {
	  quality = atoi (optarg);
	  switch (quality) {
	  case 1:
	      dv_dec->quality = DV_QUALITY_FASTEST;
	      break;
	  case 2:
	      dv_dec->quality = DV_QUALITY_AC_1;
	      break;
	  case 3:
	      dv_dec->quality = DV_QUALITY_COLOR | DV_QUALITY_DC;
	      break;
	  case 4:
	      dv_dec->quality = DV_QUALITY_COLOR | DV_QUALITY_AC_1;
	      break;
	  case 5:
  	      dv_dec->quality = DV_QUALITY_BEST;
	      break;
	  default:
	      break;
	  }
      }
      break;


      case 'f':
		input_open = xdvshow_prepare_file;
		input_func = xdvshow_read_shm;
		audio_input_func = xdvshow_read_audio_shm;
		input_close = xdvshow_close_file;
		open_option = optarg;
		flags_file = 1;
		break;

      case '4':
        rtp_param.s_addr.ss_family = (sa_family_t) AF_INET;
        break;

#ifdef	ENABLE_INET6
      case '6':
        rtp_param.s_addr.ss_family = (sa_family_t) AF_INET6;
        break;
#endif	/* ENABLE_INET6 */

      case 'P':
        rtp_param.port = atoi(optarg);
		printf("Port: %d\n", rtp_param.port);
        if (rtp_param.port < 1) {
          printf("Invalid port number : %d\n", rtp_param.port);
          return(EXIT_FAILURE);
        }
        break;

      case 'j':
		flags_mcast = 1;
		rtp_param.multi_addr_str = optarg;
		break;

      case 'A':
		flags_use_audio = 1;
		audio_dev = optarg;
		break;
		
      case 'r':
		flags_reflector = 1;
		reflector_address = optarg;
		break;

      case 'z':
		if (flags_stereo == 0) {
			flags_zoom = 1;
		} else {
			printf("Zoom not enabled with stereo flag.\n");
		}
		break;

      case 'Z':
		flags_scale = 1;
		break;

      case 'S':
		flags_stereo = 1;
		if (flags_zoom != 0) {
			flags_zoom = 0;
			printf("Zoom not enabled with stereo flag.\n");
		}
		break;

      case 'C':
		window_x = (int) strtod(optarg, &opt_tmp);
		window_y = (int) strtod(++opt_tmp, NULL);
		if ((window_x < 0) || (window_y < 0)) {
			show_usage(argv[0]);
			return(EXIT_FAILURE);
		}
		break;

      case 'R':
		window_w = (int) strtod (optarg, &opt_tmp);
		window_h = (int) strtod (++opt_tmp, NULL);
		opt_tmp = NULL;
		if ((window_w < 1) || (window_h < 1)) {
			show_usage (argv[0]);
			return (EXIT_FAILURE);
		}
		break;

      case 'W':
		window_id = strtoul (optarg, NULL, 10);
		if (window_id < 1) {
			show_usage (argv[0]);
			return (EXIT_FAILURE);
		}
		
		/* XXX - Hack to get SDL to use GTK window 
		         Is xdvshow supposed to do such things? */
		{
			char SDL_windowhack[32];
			sprintf (SDL_windowhack, "SDL_WINDOWID=%lu", window_id);
			putenv (SDL_windowhack);
		}

		printf ("SDL_WINDOWID = %s\n\n", getenv ("SDL_WINDOWID"));
		break;

      case 'I':
		flags_half_size = 1;
		break;

      case 'H':
		flags_half_framerate = 1;
		break;

      case 'F':
		flags_fullscreen = 1;
		break;

#if (LINUX && HAVE_LIBRAW1394) || FREEBSD_5
      case 'i':
		input_open = xdvshow_capture_raw;
		input_func = xdvshow_read_shm;
		audio_input_func = xdvshow_read_audio_shm;
		input_close = NULL;
		open_option = NULL;
		break;
#endif /* (LINUX && HAVE_LIBRAW1394) || FREEBSD_5 */
	
#if (HAVE_SDL || HAVE_SDL11)
      case 's':
		flags_sdl = 1;
		flags_x11 = 0;
		break;

      case 'o':
		flags_sdlnooverlay = 1;
		break;
#endif /* HAVE_SDL */
	
      case 'x':
		flags_x11 = 1;
		flags_sdl = 0;
		flags_fullscreen = 0;
		break;

      case 'd':
		flags_dump = 1;
#ifdef __USE_LARGEFILES64
		dump_fd = open("dump.dv", O_RDWR | O_CREAT | O_LARGEFILE, 00644);
#else
		dump_fd = open("dump.dv", O_RDWR | O_CREAT, 00644);
#endif /* __USE_LARGEFILES64 */
	
		if (dump_fd < 0) {
		perror("open dump.dv");
		flags_dump = 0;
		}
		break;

      case 'D':
		flags_deinterlace = 1;
		break;

      case 'h':
      default:
        show_usage(argv[0]);
        return(EXIT_FAILURE);
        break;

    }
  }

#ifdef DEBUG
  if (flags_file) {
    gettimeofday(&currtime, NULL);
    DPRINT("Start: %ld : %ld\n", currtime.tv_sec, currtime.tv_usec);
  }
#endif

  if (flags_file){
    gettimeofday(&currtime, NULL);
    frametime_usec = currtime.tv_usec + frametimeadd_usec;
  }

  pthread_attr_init(&bigstack);
  pthread_attr_setstacksize(&bigstack, (size_t) THREAD_STACKSIZE);

  /* We need to initialize semaphores first to avoid timing problems */
  if(xdvshow_semaphore_init() != 0) {
  	perror("Failed to initialize semaphores!");
	dv_decoder_free(dv_dec);
	/* XXX - Some other deallocations are probably supposed to be here! */
	exit(EXIT_FAILURE);
  }

  if (flags_reflector) {
    memset(&ping_thread, 0, sizeof(ping_thread));
    if((ret = pthread_create(&ping_thread, NULL, xdvshow_ping_reflector, (void *)reflector_address))) {
		perror("ping thread creation");
		assert(ret==0);
	}
  }

  memset(&read_thread, 0, sizeof(read_thread));
  if (input_open == xdvshow_prepare_rtp) {
	if ((ret = pthread_create(&read_thread, &bigstack, xdvshow_prepare_rtp, (void *)open_option))) {
		perror("xdvshow_prepare_rtp thread creation");
		assert(ret==0);
	}
  }
  
  pthread_attr_destroy(&bigstack);
  if (input_open == xdvshow_prepare_file) {
	if((ret = pthread_create(&read_thread, NULL, xdvshow_prepare_file, (void *)open_option))) {
		perror("xdvshow_prepare_file thread creation");
		assert(ret==0);
	}
  }
  
#if (LINUX && HAVE_LIBRAW1394) || FREEBSD_5
  if (input_open == xdvshow_capture_raw) {
	if((ret = pthread_create(&read_thread, NULL, xdvshow_capture_raw, (void *)open_option))) {
		perror("xdvshow_capture_raw thread creation");
		assert(ret==0);
	}
  }
#endif /* (LINUX && HAVE_LIBRAW1394) || FREEBSD_5 */

  DPRINT("Read thread started\n");

  while ((flags_format_normal == 0) && (flags_format_wide == 0)) {
    
    DPRINT("Reading DV frames for 16:9/4:3 detection.\n");
    ret = (*input_func)(&dvframe);
    if (ret < 0) {
      kill(getpid(), SIGINT);
      return(EXIT_FAILURE);
    }

    DPRINT("Detecting 16:9/4:3 formats.\n");

    dv_parse_header(dv_dec, dvframe);

    if (dv_format_wide(dv_dec)) {
      flags_format_normal = 1;
      fprintf (stderr, "Detected 16:9 format\n");
    }
    if (dv_format_normal(dv_dec)) {
      flags_format_wide = 1;
      fprintf (stderr, "Detected 4:3 format\n");
    }

    fprintf(stderr, "Audio is %.1f kHz, %d bits quantization, %d channels, emphasis %s\n",
	            (float)dv_dec->audio->frequency / 1000.0,
		    dv_dec->audio->quantization,
		    dv_dec->audio->num_channels,
		    (dv_dec->audio->emphasis ? "on" : "off"));

    if ((flags_format_normal != 0) && (flags_format_wide != 0)) {
      fprintf (stderr, "Error detecting 16:9/4:3 video format. Trying again...\n");
      flags_format_normal = 0;
      flags_format_wide = 0;
    }

    if (input_unlock_func) {
      (*input_unlock_func)();
    }
  }

  dv_format_type = dv_dec->system;
   switch (dv_format_type) {
    case e_dv_system_525_60:
      fprintf (stderr, "Detected NTSC format\n");
      break;
    case e_dv_system_625_50:
      fprintf (stderr, "Detected PAL format\n");
      break;
	default:
      fprintf (stderr, "Error detecting NTSC/PAL format...\n");
    }


  if (flags_use_audio) {

    ret = xdvshow_init_audio(audio_dev, dv_dec->audio);
    if (ret < 0) {
      fprintf(stderr, "Audio init failed. Audio disabled.\n");
      flags_use_audio = 0;

      goto audio_init_failed;
    }

    for (i = 0; i < 4; i++) {
      audio_buffers[i] = malloc(DV_AUDIO_MAX_SAMPLES * sizeof(u_int16_t));

      if (audio_buffers[i] == NULL) {
        fprintf(stderr, "Failed to init audio.\n");
        flags_use_audio = 0;
      }
	  else if (memset (audio_buffers[i], 0, DV_AUDIO_MAX_SAMPLES * sizeof(u_int16_t)) == NULL) {
		  perror("memset of audio_buffers[i] failed");
		  assert(0);
	  }
    }
  }

audio_init_failed:

	if(memset(&video_thread, 0, sizeof(video_thread)) == NULL) {
		perror("memset of video_thread failed");
		assert(0);
	}
	ret = pthread_create(&video_thread, NULL, video_thread_func, (void *)dv_dec);
	if(ret != 0) {
		perror("video_thread creation failed");
	}

	audio_thread_func((void *)dv_dec);

	kill(getpid(), SIGINT);

	return(EXIT_SUCCESS);
}
/** \brief This fuction is called as video displaying thread.
 *
 *  At first the function creates a new window and then repeteadly checks for
 *  any pedning events (when using libSDL), fills a videodata buffer, decodes it
 *  and displays via SDL or X11.
 *  \param *data is actually initialized dv_decoder_t structure dv_dec
 *
 *  \return NULL in all cases. This thread should be cancelled by sigint signal handling.
 */

void *
video_thread_func(void *data)
{
  int drop_frame = 0;
  int ret = 0;

  dv_decoder_t *dv_dec = NULL;

  struct x_params param;

  memset(&param, 0, sizeof(param));

  dv_dec = (dv_decoder_t *)data;

  param.dv_format_type = dv_format_type;

#if (HAVE_SDL || HAVE_SDL11)
  if (flags_sdl) {	
    ret = xdvshow_open_window(prog_name, &param);
  }
#endif /* HAVE_SDL */
  if (flags_x11) {
    ret = xdvshow_x11_open_window(prog_name, &param);
  }
  if (ret < 0) {
    printf("failed to open window\n");
    kill(getpid(), SIGINT);
    return(NULL);
  }

  while (1) {
#if (HAVE_SDL || HAVE_SDL11)
    if (flags_sdl) {
      xdvshow_SDL_handle_events();
    }
#endif /* HAVE_SDL */
    if (flags_x11) {
      xdvshow_x11_handle_events();
    }

    (*input_func)(&videodata);

    if (flags_dump) {
      ret = write(dump_fd, videodata, 144000);
      DPRINT("Videodata written: %d\n", ret);
      if(ret == -1) {
		  perror("writing videodata failed");
      }
    }

    pthread_testcancel();
    if ((!flags_half_framerate) || ((flags_half_framerate) && (!drop_frame))) {
      dv_decode_full_frame(dv_dec,
			   videodata,
			   param.decode_format,
			   param.pixels,
			   param.pitches);

      pthread_testcancel();
#if (HAVE_SDL || HAVE_SDL11)
      if (flags_sdl) {
	xdvshow_render();
      }
#endif /* HAVE_SDL */
      if (flags_x11) {
	xdvshow_x11_render();
      }
      DPRINT("DISPLAYED\n");
      pthread_testcancel();
      
      drop_frame = 1;
    
    }
    else if (flags_half_framerate) {
      drop_frame = 0;
    }

    if (input_unlock_func) {
      (*input_unlock_func)();
    }
	pthread_testcancel();

    if (flags_file){
      gettimeofday(&currtime, NULL);
      framedelay_usec = frametime_usec - currtime.tv_usec;
      if (framedelay_usec < 0 || framedelay_usec > 40000){
		DPRINT("Machine too slow or timing is queer somehow :(\n");
		DPRINT("currtime_usec: %ld, frametime_usec: %ld, framedelay_usec: %ld\n", 
					currtime.tv_usec, frametime_usec, framedelay_usec);
      } else {
		usleep(framedelay_usec);
		DPRINT("currtime_sec: %ld\n", currtime.tv_sec);
		DPRINT("currtime_usec: %ld, frametime_usec: %ld, framedelay_usec: %ld\n", 
				currtime.tv_usec, frametime_usec, framedelay_usec);
      }
      gettimeofday(&currtime, NULL);
      frametime_usec = currtime.tv_usec + frametimeadd_usec;
      if (frametime_usec > 1000000){
	frametime_usec = frametime_usec - 1000000;
      }
    }
  }

  return(NULL);
}

/** \brief This fuction is called as audio playing thread.
 *
 *  If audio is not selected by command line parameter this function if just an endless loop.
 *  In second case the function fills the audioframe buffer, calls a decode function and a play
 *  function.
 *  
 *  \param *data is not used at all
 *
 *  \return allways NULL. This thread should be cancelled by sigint signal handling.
 */

void *
audio_thread_func(void *data)
{
  int ret = 0;

  if (flags_use_audio) {
    while (audio_running) {
      (*audio_input_func)(&audioframe);

      dv_decode_full_audio(dv_dec, audioframe, audio_buffers);
      ret = xdvshow_play_audio(dv_dec->audio, audio_buffers);
      if (ret < 0) {
	break;
      }

      if (audio_unlock_func) {
	(*audio_unlock_func)();
      }
    }
  } else {
    while (1) {
      sleep(30);
    }
  }

  return(NULL);
}

