#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#ifndef S_SPLINT_S
#include <net/if.h>
#endif
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <assert.h>

#include <libdv/dv_types.h>
#include <semaphore.h>

#include "RTP.h"

#include "xdvshow-const.h"

#include "xdvshow-shm.h"
#include "xdvshow-flags.h"
#include "xdvshow-rtp.h"
#include "xdvshow-defs.h"

static int _xdvshow_prepare_socket __P((struct xdvshow_rtp_param *));
static int _xdvshow_prepare_multicast __P((struct xdvshow_rtp_param *));
static int _xdvshow_recv_dvrtp __P((void));

#ifdef DEBUG
unsigned long int received = 0;
unsigned long int packets = 0;
#endif /* DEBUG */

/** \brief periodically sends a dummy packet to the packet reflector.
 *
 * The dummy packet is sent to <tt>rtp_param.port + 1</tt>.
 *
 * \param *data is actually hostname (char *) with the packet reflector's hostname.
 * \return void.
 */

void *
xdvshow_ping_reflector(void *data)
{
  struct hostent *host = NULL;
  char *hostname = data;
  struct sockaddr_in reflector_addr;

  int sock;

  /* Code llama :-) */

  printf("Trying reflector: %s/%d\n", hostname, rtp_param.port);

  if (!(host = gethostbyname(hostname))) {
    herror(hostname);
    return((void *)1);
  }

  reflector_addr.sin_family = (sa_family_t) AF_INET;
  reflector_addr.sin_addr.s_addr = INADDR_ANY;
  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("socket()");
	free(host->h_name);
	free(host->h_aliases);
	free(host->h_addr_list);
	free(host);
    return((void *)2);
  }
  reflector_addr.sin_port = htons(rtp_param.port + 1);
  if (bind(sock, (struct sockaddr *) &reflector_addr, sizeof(reflector_addr))) {
    perror("bind()");
	free(host->h_name);
	free(host->h_aliases);
	free(host->h_addr_list);
	free(host);
    return((void *)3);
  }

  reflector_addr.sin_addr = **(struct in_addr **) host->h_addr_list;

  printf("Using reflector: %s/%d\n", hostname, rtp_param.port);

  while (1) {
    if(sendto(sock, "", 1, 0, (struct sockaddr *) &reflector_addr, sizeof(reflector_addr)) == -1) {
		perror("reflector ping sendto");
		assert(0);
	}
    sleep(REFLECTOR_PING_INTERVAL);
  }
}

/** \brief Public function that handles the input from a network.
 *
 * \return This function is never supposed to come back. -1 if the function fails for some reason.
 */

void *
xdvshow_prepare_rtp(void *data)
{
  int ret;

  struct xdvshow_rtp_param *p_rtp_param = data;

  if (p_rtp_param == NULL) {
    return((void *)-1);
  }

  ret = _xdvshow_alloc_shm();
  if (ret < 0) {
    return((void *)-1);
  }

  ret = _xdvshow_prepare_socket(p_rtp_param);

  if (p_rtp_param->multi_addr_str != NULL) {
     ret = _xdvshow_prepare_multicast(p_rtp_param);
     if (ret < 0) { 
       kill(_xdvshow_shm.parent_pid, SIGINT);
       exit(EXIT_FAILURE);
     }
   }

  ret = _xdvshow_attach_shm();
  assert(ret == 1);

  DPRINT("Calling xdvshow_recv_dvrtp.\n");
  /* THIS IS NOT SUPPOSED TO COME BACK */
  if(_xdvshow_recv_dvrtp()) {
  	DPRINT("The _xdvshow_recv_dvrtp is supposed to never return.\n");
  	exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS); /* XXX */
}

/** \brief creates a communication socket and returns a descriptor.
 *
 * \return 1 on success. -1 if the function fails for some reason.
 */

static int
_xdvshow_prepare_socket(struct xdvshow_rtp_param *rtp_param)
{
  int soc;

  struct sockaddr_in *s_in = (struct sockaddr_in *)&rtp_param->s_addr;
#ifdef	ENABLE_INET6
  struct sockaddr_in6 *s_in6 = (struct sockaddr_in6 *)&rtp_param->s_addr;
#endif	/* ENABLE_INET6 */

#ifdef	ENABLE_INET6
  if (rtp_param->s_addr.ss_family == AF_INET) {
#ifndef	NO_SS_LEN
    s_in->sin_len = sizeof(struct sockaddr_in);
#endif	/* NO_SS_LEN */
    s_in->sin_port = htons(rtp_param->port);
    s_in->sin_addr.s_addr = htonl(INADDR_ANY);
  } else if (rtp_param->s_addr.ss_family == AF_INET6) {
#ifndef	NO_SS_LEN
    s_in6->sin6_len = sizeof(struct sockaddr_in6);
#endif	/* NO_SS_LEN */
    s_in6->sin6_port = htons(rtp_param->port);
  } else {
    printf("Unknown family : %d\n", rtp_param->s_addr.ss_family);
    return(-1);
  }
#else	/* ENABLE_INET6 */
#ifndef	NO_SS_LEN
  s_in->sin_len = sizeof(struct sockaddr_in);
#endif	/* NO_SS_LEN */
  s_in->sin_port = htons(rtp_param->port);
  s_in->sin_addr.s_addr = htonl(INADDR_ANY);
#endif	/* ENABLE INET6 */

#ifdef DEBUG
  if(rtp_param->s_addr.ss_family == PF_INET6) {
    DPRINT("Preparing IPv6 socket.\n");
  }
  else if (rtp_param->s_addr.ss_family == PF_INET) {
    DPRINT("Preparing IPv4 socket.\n");
  }
  else {
    DPRINT("Preparing niether IPv4 nor IPv6 socket.\n");
    assert(0);
  }
#else /* DEBUG */
  printf("Using %s socket.\n", rtp_param->s_addr.ss_family == PF_INET6 ? "IPv6" : "IPv4" );
#endif /* DEBUG */
  soc = socket(rtp_param->s_addr.ss_family, SOCK_DGRAM, 0);
  if (soc < 0) {
    perror("socket");
    return(-1);
  }

#ifdef DEBUG
  DPRINT("Binding socket.\n");
#endif /* DEBUG */
#ifndef	NO_SS_LEN
  if (bind(soc, (struct sockaddr *)&rtp_param->s_addr, rtp_param->s_addr.ss_len) < 0) {
#else	/* NO_SS_LEN */
  if (bind(soc, (struct sockaddr *)&rtp_param->s_addr, sizeof(rtp_param->s_addr)) < 0) {
#endif	/* NO_SS_LEN */
    perror("bind");
    return(-1);
  }

  _xdvshow_shm.soc = soc;
#ifdef DEBUG
  DPRINT("Using socket fd=%d.\n", _xdvshow_shm.soc);
#endif /* DEBUG */

  return(1);
}

/** \brief converts the multicast address into binary data in network byte order.
 *
 * \param *rtp_param (struct xdvshow_rtp_param)
 * \return 1 on success. -1 if the address is invalid.
 */

static int
_multicast_str2addr(struct xdvshow_rtp_param *rtp_param)
{
#ifdef  ENABLE_INET6
  struct addrinfo hints, *res;

  memset(&hints, 0, sizeof(hints));

  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_protocol = IPPROTO_UDP;

  if (getaddrinfo(rtp_param->multi_addr_str, NULL, &hints, &res) != 0) {
    perror("getaddrinfo");
    return(-1);
  }

  memcpy(&rtp_param->multi_s_addr, res->ai_addr, res->ai_addrlen);
#else   /* ENABLE_INET6 */
  struct sockaddr_in *s_in = (struct sockaddr_in *)&rtp_param->multi_s_addr;

  s_in->sin_addr.s_addr = inet_addr(rtp_param->multi_addr_str);
  if (s_in->sin_addr.s_addr == -1) {
    return(-1);
  }
#endif  /* ENABLE_INET6 */

  return(1);
}

/** \brief checks if the given address is a correct multicast address.
 *
 * \param *rtp_param (struct xdvshow_rtp_param)
 * \return 1 on success. -1 if the function fails for some reason.
 */

static int
_chk_multicast_addr(struct xdvshow_rtp_param *rtp_param)
{
  struct sockaddr_in *s_in = (struct sockaddr_in *)&rtp_param->multi_s_addr;
#ifdef  ENABLE_INET6
  struct sockaddr_in6 *s6_in = (struct sockaddr_in6 *)&rtp_param->multi_s_addr;
#endif  /* ENABLE_INET6 */

  switch (rtp_param->multi_s_addr.ss_family) {
    case AF_INET:
      if (!IN_MULTICAST(ntohl(s_in->sin_addr.s_addr))) {
        printf("ERROR : [%s] is not multicast address\n",
               rtp_param->multi_addr_str);
        return(-1);
      }
      break;

#ifdef  ENABLE_INET6
    case AF_INET6:
      if (!IN6_IS_ADDR_MULTICAST(&s6_in->sin6_addr)) {
        printf("ERROR : [%s] is not multicast address\n",
               rtp_param->multi_addr_str);
        return(-1);
      }
      break;
#endif  /* ENABLE_INET6 */

    default:
      printf("ERROR : unknown multicast addr family : %d\n",
             rtp_param->multi_s_addr.ss_family);
      return(-1);
      break;
  }

  return(1);
}


/** \brief checks the multicast address family.
 *
 * \param *rtp_param (struct xdvshow_rtp_param)
 * \return 1 on success. -1 if the function fails for some reason.
 */

static int
_chk_multicast_family(struct xdvshow_rtp_param *rtp_param)
{
  if (rtp_param->s_addr.ss_family !=
        rtp_param->multi_s_addr.ss_family) {
    printf("Protocol family of DV/RTP socket and multicast join address mismatch\n");

    switch (rtp_param->s_addr.ss_family) {
      case AF_INET:
        printf("DV/RTP socket family   : IPv4 (AF_INET)\n");
        break;

#ifdef ENABLE_INET6
      case AF_INET6:
        printf("DV/RTP socket family   : IPv6 (AF_INET6)\n");
        break;
#endif  /* ENABLE_INET6 */

      default:
        break;
    }
    switch (rtp_param->multi_s_addr.ss_family) {
      case AF_INET:
        printf("multicast join address : IPv4 (AF_INET)\n");
        break;

#ifdef ENABLE_INET6
      case AF_INET6:
        printf("multicast join address : IPv6 (AF_INET6)\n");
        break;
#endif  /* ENABLE_INET6 */

      default:
        break;
    }

    return(-1);
  }

  return(1);
}

/** \brief is an internal function that joins mlticast group given by its address.
 *
 * \param *rtp_param (struct xdvshow_rtp_param)
 * \return 1 on success. -1 if the function fails for some reason.
 */

static int
_multicast_join(struct xdvshow_rtp_param *rtp_param)
{
  struct ip_mreq mreq;
  struct sockaddr_in *s_in = (struct sockaddr_in *)&rtp_param->multi_s_addr;
  struct sockaddr_in *base_s_in = (struct sockaddr_in *)&rtp_param->s_addr;
#ifdef  ENABLE_INET6
  struct ipv6_mreq mreq6;
  struct sockaddr_in6 *s6_in = (struct sockaddr_in6 *)&rtp_param->multi_s_addr;

  memset(&mreq6, 0, sizeof(mreq6));
#endif  /* ENABLE_INET6 */

  memset(&mreq, 0, sizeof(mreq));

  switch (rtp_param->multi_s_addr.ss_family) {
    case AF_INET:
      mreq.imr_multiaddr.s_addr = s_in->sin_addr.s_addr;
      mreq.imr_interface.s_addr = base_s_in->sin_addr.s_addr;

      if (setsockopt(_xdvshow_shm.soc, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                     (char *)&mreq, sizeof(mreq)) < 0) {
        perror("ip_add_membership");
        printf("multi addr     : %s\n", inet_ntoa(mreq.imr_multiaddr));
        printf("interface addr : %s\n", inet_ntoa(mreq.imr_interface));
      }
      break;

#ifdef  ENABLE_INET6
    case AF_INET6:
      memcpy((char *)&mreq6.ipv6mr_multiaddr,
             (char *)&s6_in->sin6_addr,
             sizeof(s6_in->sin6_addr));

      if (rtp_param->multi_ifname == NULL) {
        printf("ERROR : multicast interface required\n");
        printf("  ex) xdvshow -j ff02::ab:cd -M fxp0\n");
        return(-1);
      }
      mreq6.ipv6mr_interface = if_nametoindex(rtp_param->multi_ifname);

#ifdef  IPV6_ADD_MEMBERSHIP
      if (setsockopt(_xdvshow_shm.soc, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
                     (char *)&mreq6, sizeof(mreq6)) < 0) {
#else   /* IPV6_ADD_MEMBERSHIP */
      if (setsockopt(_xdvshow_shm.soc, IPPROTO_IPV6, IPV6_JOIN_GROUP,
                     (char *)&mreq6, sizeof(mreq6)) < 0) {
#endif  /* IPV6_ADD_MEMBERSHIP */
        perror("ip_add_membership");
      }
      break;
#endif  /* ENABLE_INET6 */

    default:
      return(-1);
      break;
  }

  return(1);
}

/** \brief converts the given address, checks if the address is 
 * correct and then joins the multicast group.
 *
 * \param *rtp_param (struct xdvshow_rtp_param)
 * \return 1 on success. -1 otherwise.
 */

static int
_xdvshow_prepare_multicast(struct xdvshow_rtp_param *rtp_param)
{
  int ret;

  ret = _multicast_str2addr(rtp_param);
  if (ret < 0) {
    return(-1);
  }

  ret = _chk_multicast_addr(rtp_param);
  if (ret < 0) {
    return(-1);
  }

  ret = _chk_multicast_family(rtp_param);
  if (ret < 0) {
    return(-1);
  }

  ret = _multicast_join(rtp_param);
  if (ret < 0) {
    return(-1);
  }

  return(1);
}

/** \brief internal function that acquires the DV stream from a network.
  *
  * \return 0 on success, -1 on failure
  */ 
static int
_xdvshow_recv_dvrtp()
{
  int n;
  int i;

  u_long *recvbuf;
  u_long *recvbuf_ptr;

  u_char *dvdata;
  u_char *audio_dvdata;

#ifdef DEBUG
  int sem_value = 0;
  int frame_length = 0;
  int iDebugSemVal;

  int j;

  int dseq_array[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
#endif /* DEBUG */

  unsigned int dbn, dseq, seq, sct;

  unsigned int dseq_max;
  unsigned int dseq_stream_max = 0; 
  unsigned int dseq_stream_max_count = 0;

  recvbuf = (u_long *) malloc(sizeof(u_long) * MAX_PKT_LEN);

  switch (dv_format_type) {
    case e_dv_system_525_60:
      dseq_max = DSEQ_MAX_NTSC;
      break;

    case e_dv_system_625_50:
      dseq_max = DSEQ_MAX_PAL;
      break;

    default:
      free(recvbuf);
      return(-1);
  }

  dvdata = _xdvshow_shm.shm_frame->frame_buf->data;
  audio_dvdata = _xdvshow_shm.audio_shm_frame->frame_buf->data;

  while (1) {

  DPRINT("Receiving RTP.\n");

  n = read(_xdvshow_shm.soc, (char *)recvbuf, sizeof(u_long) * MAX_PKT_LEN);

  if (n < 1) {
    return(-1);
  }
	

#ifdef DEBUG
  received += n;
  frame_length +=n;
  packets++;
  DPRINT("RECEIVED %ldb; Received packets: %ld\n", received, packets);
#endif /* DEBUG	*/
  	    
  /* XXX  - recvbuf -> u_char */
  assert(sizeof(rtp_hdr_t) % 4 == 0);
  for (i=0; i<(n - sizeof(rtp_hdr_t))/80; i++) {
    DPRINT_BARE(".");
    recvbuf_ptr = recvbuf + (sizeof(rtp_hdr_t)+80*i)/4;
    assert((char *)recvbuf_ptr < (char *)recvbuf + n);
    dbn  = (ntohl(*recvbuf_ptr) >> 8) & 0xff;
    dseq = (ntohl(*recvbuf_ptr) >> 20) & 0xf;
    seq = (ntohl(*recvbuf_ptr) >> 24) & 0xf;
    sct  = (ntohl(*recvbuf_ptr) >> 29) & 0x7;

    if (sct > 4) {
	DPRINT_BARE("\n");
	DPRINT("Data error: SCT > 4! sct = %d\n", sct);
	continue;
    }


    /* Hard-testing just DSEQ_MAX_PAL is to facilitate on-the-fly change from
     * NTSC to PAL without restarting application.
     */
	  if (dseq > dseq_max) {
		DPRINT_BARE("\n");
		DPRINT("Data error: DSEQ > DSEQ_MAX! dseq = %d, dseq_max = %d\n", dseq, dseq_max);
	  }
	  if (dseq > DSEQ_MAX_PAL) {
		DPRINT_BARE("\n");
		DPRINT("Data error: DSEQ > DSEQ_MAX_PAL! dseq = %d, DSEQ_MAX_PAL = %d\n", dseq, DSEQ_MAX_PAL);
		continue;
	  }

      /* XXX This is dirty temporary solution for switching PAL/NTSC 
	 on the fly. It measures dseq maximum in window of 255 and uses it as
         dseq_max.

	 XXX There is something rotten in dseq_max measuring. dseq_max shouldn't
	 change in the middle of the stream (using dvsend to send the stream).
       */
	  if (flags_mcast) {
	  
	      if (dseq_stream_max_count >= 255) {
		if (dseq_stream_max != dseq_max && (dseq_stream_max == DSEQ_MAX_PAL || dseq_stream_max == DSEQ_MAX_NTSC) ) {
			    dseq_max = dseq_stream_max;
		    }
		    dseq_stream_max_count = 0;
		    dseq_stream_max = 0;
	      }
	      else {
		    if (dseq_stream_max < dseq) {
			    dseq_stream_max = dseq;
		    }
		    ++dseq_stream_max_count;
	      }
	  
	  }

#ifdef DEBUG
    if ((dseq > 0 ) && (dseq <= dseq_max)){
      dseq_array[dseq] = 1;
    }
#endif /* DEBUG */


      switch (sct) {

        case SCT_HEADER:
	  if((dseq * 150 *80) <= (DVFRAME_MAX_SIZE - 80))
            memcpy(&dvdata[dseq * 150 * 80], recvbuf_ptr, 80);
          else {
		DPRINT_BARE("\n");
		DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 * 80) - DVFRAME_MAX_SIZE + 80), dbn);
          }
          break;

        case SCT_SUBCODE:
	  if(dbn > 1) {
		DPRINT_BARE("\n");
		DPRINT("Data error: DBN > 1\ndbn = %d\n", dbn);
		break;
	  }
	  if((dseq * 150 + 1 + dbn) *80 <= (DVFRAME_MAX_SIZE - 80))
            memcpy(&dvdata[(dseq * 150 + 1 + dbn) * 80], recvbuf_ptr, 80);
          else {
		DPRINT_BARE("\n");
		DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 1 + dbn)* 80 - DVFRAME_MAX_SIZE + 80), dbn);
          }
          break;


        case SCT_VAUX:
	  if(dbn > 2) {
		DPRINT_BARE("\n");
		DPRINT("Data error: DBN > 2\ndbn = %d\n", dbn);
		break;
	  }
	  if((dseq * 150 + dbn + 3)*80 <= (DVFRAME_MAX_SIZE - 80))
            memcpy(&dvdata[(dseq * 150 + dbn + 3) * 80], recvbuf_ptr, 80);
          else {
		DPRINT_BARE("\n");
		DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d, dbn = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 6 + dbn * 16)*80 - DVFRAME_MAX_SIZE + 80), dseq, dbn);
          }
          break;

        case SCT_AUDIO:
	  if(dbn > 8) {
		DPRINT_BARE("\n");
		DPRINT("Data error: DBN > 8\ndbn = %d\n", dbn);
		break;
	  }
	  DPRINT_BARE("*");
	  if((dseq * 150 + 6 + dbn * 16)*80 <= (DVFRAME_MAX_SIZE - 80))
            memcpy(&dvdata[(dseq * 150 + 6 + dbn*16) * 80], recvbuf_ptr, 80);
          else {
		DPRINT_BARE("\n");
		DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d, dbn = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 6 + dbn * 16)*80 - DVFRAME_MAX_SIZE + 80), dseq, dbn);
		break;
          }
	  if (flags_use_audio) {
	    DPRINT_BARE("(*)");
	    if((dseq * 150 + 6 + dbn * 16) * 80  <= (DVFRAME_MAX_SIZE - 80))
	      memcpy(&audio_dvdata[(dseq * 150 + 6 + dbn * 16) * 80], recvbuf_ptr, 80);
	    else {
		  DPRINT_BARE("\n");
		  DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d, dbn = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 6 + dbn * 16)*80 - DVFRAME_MAX_SIZE + 80), dseq, dbn);
		  break;
	    }

	  /* XXX This is very nasty and prevents from going from PAL to NTSC
		     unless dseq_max is adjusted properly since dseq_max for PAL is never
		     reached in NTSC.
		   */
	  if (dseq == dseq_max && dbn == 8) {
#ifdef DEBUG
            if(sem_getvalue(&audio_ready, &iDebugSemVal)) {
			  perror("Unable to get value of audio_ready semaphore!");
			  exit(EXIT_FAILURE);
	    }
	    DPRINT("semaphore audio_ready value: %d\n", iDebugSemVal);

	    if(sem_getvalue(&audio_empty, &iDebugSemVal)) {
			  perror("Unable to get value of audio_empty semaphore!");
			  exit(EXIT_FAILURE);
	    }
	    DPRINT("semaphore audio_empty value: %d\n", iDebugSemVal);
#endif /* DEBUG */

  /* The sem_post and sem_wait should happen if 16:9/4:3 detection is
   * finished. This was the problem with xdvshow deadlocking after error
   * in detection of 16:9/4:3 formats on FreeBSD. At least I hope so. */
	    if ( !( ((flags_format_normal == 0) && (flags_format_wide == 0))
   	         || ((flags_format_normal != 0) && (flags_format_wide != 0)) ) )
              {
	        sem_post(&audio_ready);
	        ts_sem_wait(&audio_empty);
	      }

              _xdvshow_shm.audio_shm_frame = _xdvshow_shm.audio_shm_frame->next;
	      audio_dvdata = _xdvshow_shm.audio_shm_frame->frame_buf->data;
	    }
	  }
	    
          break;

        case SCT_VIDEO:
	  if(dbn > 134) {
		  DPRINT_BARE("\n");
		  DPRINT("Data error: DBN > 134\ndbn = %d\n", dbn);
		  break;
	  }
	  if((dseq * 150 + 7 + dbn + dbn/15) * 80  > (DVFRAME_MAX_SIZE - 80)) {
		  DPRINT_BARE("\n");
		  DPRINT("memcpy dropped - dvdata size %d exceeded by %d\ndseq = %d, dbn = %d\n", DVFRAME_MAX_SIZE, ((dseq * 150 + 6 +7 + dbn + dbn/15)*80 - DVFRAME_MAX_SIZE + 80), dseq, dbn);
	  }
          memcpy(&dvdata[(dseq * 150 + 7 + dbn + dbn/15) * 80],
                 recvbuf_ptr,
                 80);

	  if (dseq == dseq_max && dbn == 134) {
            _xdvshow_shm.shm_frame->frame_buf->lock = DVFRAME_DATA_READY;

#ifdef DEBUG
	  DPRINT("RECEIVED frame no. %d\n", sem_value++);
	  DPRINT("RECEIVED frame length: %d\n", frame_length);

	  /* Sequence numbers checking */
	  DPRINT_BARE("\n\n DSEQ_ARRAY [ .");
	  for (j=1; j <= dseq_max; j++) {
	      DPRINT_BARE("%d.", dseq_array[j]);
	  }
	  DPRINT_BARE(" ]\tDSEQ_MAX: %d\n\n", dseq_max);

	  for (j=0; j <= dseq_max; j++) {
	      dseq_array[j] = 0;
	  }


	  if (sem_getvalue(&video_ready, &iDebugSemVal)) {
	    perror("Unable to get value of video_ready semaphore!");
	    exit(EXIT_FAILURE);
	  }
	  DPRINT("semaphore video_ready value: %d\n", iDebugSemVal);

	  if (sem_getvalue(&video_empty, &iDebugSemVal)) {
	    perror("Unable to get value of video_empty semaphore!");
	    exit(EXIT_FAILURE);
	  }
	  DPRINT("semaphore video_empty value: %d\n", iDebugSemVal);

	  frame_length = 0;
#endif /* DEBUG */

	  sem_post(&video_ready);
	  ts_sem_wait(&video_empty);

	  _xdvshow_shm.shm_frame = _xdvshow_shm.shm_frame->next;
	  dvdata = _xdvshow_shm.shm_frame->frame_buf->data;
	}
	break;

        default:
          break;
      }
    }
  }

  return(0);
}

