/*
 * Copyright (c) 1999, 2000, 2001, 2002 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <libdv/dv_types.h>
#include <pthread.h>

#ifdef HAVE_SDL
#include <SDL/SDL.h>
#include <SDL/SDL_syswm.h>
#endif /* HAVE_SDL */

#ifdef HAVE_SDL11
#include <SDL11/SDL.h>
#include <SDL11/SDL_syswm.h>
#endif /* HAVE_SDL11 */

#include "xdvshow-sdl.h"
#include "xdvshow-flags.h"
#include "xdvshow-defs.h"

#include <postprocess.h>

SDL_Surface *sdl_screen;
SDL_Overlay *overlay;
SDL_Overlay *dstoverlay;
SDL_Rect rect;

SDL_Event sdl_event;

#ifndef S_SPLINT_S
static Display *display;
#endif

int width, height;

unsigned int res_x = 1024, res_y = 768;  /* Actual size of root window */

extern int window_x;
extern int window_y;
extern int window_w;
extern int window_h;
extern unsigned long window_id;


pp_context_t *pp_context;
pp_mode_t *pp_mode;

/** \brief Internal helper function getting a display resolution.
 *
 * \return This function should allways finish returning 0.
 * 
 */

int Get_resolution(){

  int itemp;
  unsigned int utemp;
#ifndef S_SPLINT_S
  Window wtemp;
  Status temp;
#endif

  temp = XGetGeometry(display, DefaultRootWindow(display), &wtemp, &itemp, &itemp, &res_x, &res_y, &utemp, &utemp);

  if (XQueryExtension(display, "XINERAMA", &itemp, &itemp, &itemp)) {
    fprintf(stdout, "XINERAMA detected but not used in this version\n");
  }
  printf("Display resolution: %d x %d.\n", res_x, res_y);

  if (flags_stereo)
	res_x = res_x / 2;

  return 0;
}

/** \brief converts colorspaces between YUY2 and IYUV (I420).
 *
 * \param **src YUY2 decoded DV frame
 * \param **dst IYUV frame for further processing by deinterlacer
 *
 * \return void.
 */

void yuy2yuv(uint8_t **src, uint8_t **dst) {

  unsigned char y2y[256];
  unsigned char br2br[256];

  unsigned char *yuv;

  int x, y, i, j, j0, j1, j2, j3, k;

  for (i = 0; i < 256; i++) {
    j = 16 + i * (236-16) / 255;
#if 0
    y2y[i] = j;
    br2br[i] = 16 + i * (240-16) / 255;
#else
    y2y[i] = i;
    br2br[i] = i;
#endif
  }

  yuv = (unsigned char *)malloc(2 * overlay->w * overlay->h * sizeof(unsigned char));

  /*
  memcpy(yuv, src, 2 * overlay->w * overlay->h);
  */
  yuv = (unsigned char *)src;

  for (y = 0; y < overlay->h; y++) {
    for (x = 0; x < overlay->w; x++) {
      memcpy(dst, (unsigned char *)y2y[yuv[2 * (y * overlay->w + x) + 1]], sizeof(unsigned char));
    }
  }

  j0 = 0;
  j1 = 2 * overlay -> w;

  for (i = 0; i < 2; i++) {
    for (y = 0; y < overlay->h; y += 2) {
      j2 = 0;
      if (y%4) j2 =j0;
      j3 = j2 + j1;

      for (x = 0; x < overlay->w; x += 2) {
	j = 2 * (y * overlay->w + x + i);
	k = (yuv[j + j2] + yuv[j + j3])/2;
	memcpy(dst, (unsigned char *)br2br[k], sizeof(unsigned char));
      }
    }
  }

  free(yuv);

}


static void _dv_center_window __P((SDL_Surface *));

/** \brief SDL Event handler
 *
 * Function looks for any pending events (eg. key was pressed) and 
 * then handles it. Actually there is only a quit event implemented.
 *
 * \return 0 when finished.
 */

int
xdvshow_SDL_handle_events(void) {
  pthread_testcancel();
  while (SDL_PollEvent(&sdl_event)) {
    pthread_testcancel();
    switch (sdl_event.type) {
      case SDL_KEYDOWN:
      case SDL_KEYUP:
	
	if (!strcmp(SDL_GetKeyName(sdl_event.key.keysym.sym), "q")) {
          fprintf(stderr, "Exiting...\n");
  	  kill(0, SIGINT);
	}
	break;

      default:
	break;
    }
  }
  return(0);
}

/** \brief This function creates a new window with SDL screen inside.
 *
 * \param *window_name Name of the window.
 * \param *param The video resolution is given within this structure.
 *
 * \return 1 if there was no problem. -1 otherwise.
 */

int
xdvshow_open_window(char *window_name, struct x_params *param)
{
  char *pp_name;

  int ret;
  const SDL_VideoInfo *video_info;
  int video_bpp;

  switch (param->dv_format_type) {
    case e_dv_system_525_60:
      width = 720;
      height = 480;
      break;

    case e_dv_system_625_50:
      width = 720;
      height = 576;
      break;

    default:
      return(-1);
  }

  if (window_id == 0) {
    fprintf(stdout, "Opening connection to X server.\n");
    display = XOpenDisplay(NULL);
  }

  ret = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE);
  if (ret < 0) {
    return(-1);
  }

  video_info = SDL_GetVideoInfo();
  /*
  switch (video_info->vfmt->BitsPerPixel) {
    case 16:
    case 32:
      video_bpp = video_info->vfmt->BitsPerPixel;
      break;

    default:
      video_bpp = 16;
      break;
  }
  */
  /* If video_bpp is 0 then the current bpp mode will be used. */
  video_bpp = 0;

  if (flags_sdlnooverlay) {

    printf("Display settings: SDL HW_SURFACE disabled.\n");

    if (flags_fullscreen) {
      if (flags_zoom) {
	res_x = width;
	res_y = height;
	sdl_screen = SDL_SetVideoMode(width, height, video_bpp, SDL_SWSURFACE | SDL_FULLSCREEN);
      } else {
	Get_resolution();
	if (flags_stereo) {
	  sdl_screen = SDL_SetVideoMode(res_x, res_y, video_bpp, SDL_SWSURFACE | SDL_NOFRAME);
	} else {
	  sdl_screen = SDL_SetVideoMode(res_x, res_y, video_bpp, SDL_SWSURFACE | SDL_FULLSCREEN);
	}
      }
    } else {
      sdl_screen = SDL_SetVideoMode(width, height, video_bpp, SDL_SWSURFACE);
    }

  } else {

    if (flags_fullscreen) {
      if (flags_zoom) {
	res_x = width;
	res_y = height;
	sdl_screen = SDL_SetVideoMode(width, height, video_bpp, SDL_HWSURFACE | SDL_FULLSCREEN | SDL_DOUBLEBUF);
      } else {
	Get_resolution();
	if (flags_stereo) {
	  sdl_screen = SDL_SetVideoMode(res_x, res_y, video_bpp, SDL_HWSURFACE | SDL_NOFRAME | SDL_DOUBLEBUF);
	} else {
	  sdl_screen = SDL_SetVideoMode(res_x, res_y, video_bpp, SDL_HWSURFACE | SDL_FULLSCREEN | SDL_DOUBLEBUF);
	}
      }
    } else {
      if (flags_half_size) {
	/* TODO: flags_half_size */
        sdl_screen = SDL_SetVideoMode(width, height, video_bpp, SDL_HWSURFACE | SDL_DOUBLEBUF);
      } else {
        sdl_screen = SDL_SetVideoMode(width, height, video_bpp, SDL_HWSURFACE | SDL_DOUBLEBUF);
      }
    }

  }

  SDL_WM_SetCaption(window_name, "name");

  SDL_ShowCursor(SDL_DISABLE);

#define	DV_FOURCC_YUY2	0x32595559
#define DV_FOURCC_IYUV  0x56555949
  overlay = SDL_CreateYUVOverlay(width, height, DV_FOURCC_YUY2, sdl_screen);

  if (!overlay) {
    fprintf(stderr, "SDL_overlay initialization failed.\n");
  }

  if(flags_deinterlace) {
    dstoverlay = SDL_CreateYUVOverlay(width, height, DV_FOURCC_IYUV, sdl_screen);
  }  

  _dv_center_window(sdl_screen);

  if (flags_fullscreen) {
    if (flags_scale) {
      rect.w = res_x;
      rect.h = ((float)res_x / (float)width) * overlay->h;
	  if (rect.h > res_y) {
	    rect.h = res_y;
		rect.w = ((float)rect.h / (float)height) * width;
        rect.x = (res_x - rect.w) / 2;
        rect.y = 0;
	  }
	  else {
        rect.x = 0;
        rect.y = (res_y - rect.h) / 2;
	  }
    } else {
      rect.x = (res_x - overlay->w)/2;
      rect.y = (res_y - overlay->h)/2;
      rect.w = overlay->w;
      rect.h = overlay->h;
    }
  } else {
    rect.x = 0;
    rect.y = 0;
    rect.w = overlay->w;
    rect.h = overlay->h;
  }

  printf("Display settings: rect_top: %d x %d; rect_res: %d x %d.\n", rect.x, rect.y, rect.w, rect.h);

  param->pixels[0] = overlay->pixels[0];
  param->pixels[1] = overlay->pixels[1];
  param->pixels[2] = overlay->pixels[2];
  param->pitches[0] = overlay->pitches[0];
  param->pitches[1] = overlay->pitches[1];
  param->pitches[2] = overlay->pitches[2];

  param->decode_format = e_dv_color_yuv;

  /* Optional deinterlace init stuff */
  if (flags_deinterlace) {
    DPRINT("PP init...\n");
    pp_name = (char *)malloc(10);
    sprintf(pp_name, "lb");

    /* TODO: CPU capabilities handling */
    pp_context = pp_get_context(rect.w, rect.h, PP_CPU_CAPS_MMX);
    /* TODO: pp_name recognizing */
    pp_mode = pp_get_mode_by_name_and_quality(pp_name, PP_QUALITY_MAX);
    free(pp_name);
    DPRINT("PP init... done.\n");
  }

  return(1);
}

/** \brief closes the sdl window.
 *
 * There is also some cleanup after postprocessing (if required).
 * 
 * \return 1 after successfully finished.
 */

int
xdvshow_close_window(void)
{
  /* Optional deinterlace uinit stuff */
  if (flags_deinterlace) {
    DPRINT("PP uinit...\n");
    if (pp_mode) pp_free_mode(pp_mode);
    if (pp_context) pp_free_context(pp_context);
    DPRINT("PP uinit... done.\n");
  }

  /* It seems, that SDL eats mouse cursor sometimes. */
  SDL_ShowCursor(SDL_ENABLE);
  
  printf("Quitting SDL.\n");
  SDL_Quit();

  return(1);
}

/** \brief Function that renders the video.
 *
 * If there is any postprocessing required, the function converts the frame
 * to IYUV and then calls pp_postprocess from libpostproc.
 *
 * \return 1 after finishing successfully.
 */

int
xdvshow_render(void)
{
  int srcStride[3] = {720, 360, 360};
  int dstStride[3] = {720, 360, 360};

  pthread_testcancel();
  XFlush(display);
  pthread_testcancel();
  SDL_UnlockYUVOverlay(overlay);
  pthread_testcancel();
  if (flags_deinterlace) {
    /* TODO: overlay->pixels segfault because of overlay->pixels is stored
     * in YUY2 format and pp_postprocess needs I420 aka IYUV format*/
    yuy2yuv(overlay->pixels, dstoverlay->pixels);

    pp_postprocess(overlay->pixels, srcStride, dstoverlay->pixels, dstStride, 
                   overlay->w, overlay->h, 0, 0, pp_mode, pp_context, 0);
    pthread_testcancel();
    SDL_DisplayYUVOverlay(dstoverlay, &rect);
  } else {
    SDL_DisplayYUVOverlay(overlay, &rect);
  }
  pthread_testcancel();
  SDL_LockYUVOverlay(overlay);
  pthread_testcancel();

  return(1);
}

/** \brief Just an internal function that centers the SDL window within whole display.
 *
 * \param *screen is the initialized SDL_surface.
 * \return void
 */

static void
_dv_center_window(SDL_Surface *screen)
{
  SDL_SysWMinfo info;

  SDL_VERSION(&info.version);
  if ( SDL_GetWMInfo(&info) > 0 ) {
    int x, y;
    int w, h;
    if (info.subsystem == SDL_SYSWM_X11) {
      info.info.x11.lock_func();
      w = DisplayWidth(info.info.x11.display,
                       DefaultScreen(info.info.x11.display));
      h = DisplayHeight(info.info.x11.display,
                        DefaultScreen(info.info.x11.display));
      x = (w - screen->w)/2;
      y = (h - screen->h)/2;
      XMoveWindow(info.info.x11.display, info.info.x11.wmwindow, x, y);
      info.info.x11.unlock_func();
    }
  }
}
