/*
 * Copyright (c) 1999, 2000, 2001, 2002 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <assert.h>
#include <time.h>
#include <sys/shm.h>
#include <assert.h>
#include <errno.h>
#include <pthread.h>

#include "xdvshow-const.h"

#include "xdvshow-shm.h"
#include "xdvshow-flags.h"
#include "xdvshow-defs.h"

extern u_int32_t flags;
extern u_int32_t dv_format_type;

struct __xdvshow_shm _xdvshow_shm;

sem_t video_ready, video_empty;
sem_t audio_ready, audio_empty;

/**
 * \brief xdvshow_semaphore_init() initializes video_ready, video_empty,
 * audio_ready, and audio_empty semaphores.
 *
 * xdvshow_semaphore_init() initializes video_ready, video_empty, audio_ready,
 * and audio_empty semaphores. *_ready semaphores are initialized to 0. *_empty
 * semaphores are initialized to _xdvshow_shm.framebufs value.
 * \return 0 on success, -1 on failure
 */
int xdvshow_semaphore_init()
{
  if ((sem_init(&video_ready, 0, 0) != 0) ||
    (sem_init(&video_empty, 0, _xdvshow_shm.framebufs) != 0)) {

    perror("sem_init");
    return (-1);
  }

  if (flags_use_audio) {
    if ((sem_init(&audio_ready, 0 , 0) != 0) ||
      (sem_init(&audio_empty, 0, _xdvshow_shm.framebufs) != 0)) {

      perror("audio sem_init");
      return (-1);
    }
  }

  return(0);
}

/** \brief xdvshow_close_shm() detaches shared memory _xdvshow_shm.shm_buf.
  */
void
xdvshow_close_shm()
{
  if(shmdt(_xdvshow_shm.shm_buf)) {
	  perror("detaching shared memory _xdvshow_shm.shm_buf");
	  assert(0);
  }
}

/** \brief xdvshow_read_unlock() decrements (unlocks) video_empty semaphore
  *
  * \return always 1
  */
int
xdvshow_read_unlock()
{
  _xdvshow_shm.locked_shm_framebuf->lock = DVFRAME_READY;

  if (sem_post(&video_empty) < 0) {
    perror("video_empty sem_post");
    assert(0);
  }

  return(1);
}

/** \brief xdvshow_read_audio_unlock() decrements (unlocks) audio_empty
  * semaphore
  *
  * \return always 1
  */
int
xdvshow_read_audio_unlock()
{
  _xdvshow_shm.locked_audio_shm_framebuf->lock = DVFRAME_READY;

  if (sem_post(&audio_empty) < 0) {
    perror("audio_empty sem_post");
    assert(0);
  }

  return(1);
}

/** \brief returns the video data of the DV frame.
 *
 * Waits for the video_ready semaphore and then points the dvframe pointer
 * to the DV video buffer in the shared memory segment.
 * 
 * \param **dvframe (unsigned char) is the curently processed frame buffer.
 * \return 1 after finishing without any problems.
 */

int
xdvshow_read_shm(u_char **dvframe)
{
#ifdef DEBUG
	int iDebugSemVal;
	if(sem_getvalue(&video_ready, &iDebugSemVal)) {
		perror("Unable to get value of video_ready semaphore!");
		exit(EXIT_FAILURE);
	}
	DPRINT("semaphore video_ready value: %d\n", iDebugSemVal);
#endif /* DEBUG */

	ts_sem_wait(&video_ready);


  _xdvshow_shm.read_frame->frame_buf->lock = DVFRAME_COPYING;

  *dvframe = _xdvshow_shm.read_frame->frame_buf->data;
  _xdvshow_shm.locked_shm_framebuf = _xdvshow_shm.read_frame->frame_buf;

  _xdvshow_shm.read_frame = _xdvshow_shm.read_frame->next;

  return(1);
}

/** \brief returns the audio data of the DV frame.
 *
 * Waits for the audio_ready semaphore and then points the audioframe pointer
 * to the audio data in the shared memory segment.
 *
 * \param **audioframe (unsigned char) is the curently processedaudio buffer.
 * \return 1 after finishing without any problems.
 */

int
xdvshow_read_audio_shm(u_char **audioframe)
{
#ifdef DEBUG
	int iDebugSemVal;
	if(sem_getvalue(&audio_ready, &iDebugSemVal)) {
		perror("Unable to get value of audio_ready semaphore!");
		exit(EXIT_FAILURE);
	}
	DPRINT("semaphore audio_ready value: %d\n", iDebugSemVal);
#endif /* DEBUG */

	ts_sem_wait(&audio_ready);

  _xdvshow_shm.audio_shm_frame->next->frame_buf->lock = DVFRAME_COPYING;

  *audioframe = _xdvshow_shm.audio_shm_frame->frame_buf->data;
  _xdvshow_shm.locked_audio_shm_framebuf = _xdvshow_shm.audio_shm_frame->frame_buf;

  _xdvshow_shm.audio_shm_frame = _xdvshow_shm.audio_shm_frame->next;

  return(1);
}

/** \brief allocates a shared memory segment for video frame buffers list.
 *
 * Private function.
 *
 * \return 1 on success (XXX), -1 on failure
 */

int
_xdvshow_alloc_shm()
{
  int shmid;

  shmid = shmget(IPC_PRIVATE,
                 SHM_FRAME_BUF_NUM*sizeof(struct shm_frame_buf)*2,	/* audio and video */
                 0600);
  if (shmid < 0) {
    perror("shmget");
    return(-1);
  }

  _xdvshow_shm.shmid = shmid;
  _xdvshow_shm.framebufs = SHM_FRAME_BUF_NUM;

  return(1);
}

/** \brief allocates the video and audio buffers within the frame buffers list.
 *
 * Private function.
 * The number of buffers depends on <tt>SHM_FRAME_BUF_NUM</tt>.
 *
 * \return 1 on success (XXX), -1 on failure
 */

int
_xdvshow_attach_shm()
{
  u_long *shm_buf;

  int i;
  struct shm_frame *shm_frame = NULL;
  struct shm_frame *shm_frame_last = NULL;

  if ((shm_buf = shmat(_xdvshow_shm.shmid, 0, 0)) == (void *)-1) {
    perror("shmat");
    return(-1);
  }

  _xdvshow_shm.shm_buf = shm_buf;

  for (i=0; i<_xdvshow_shm.framebufs; i++) {
    shm_frame = (struct shm_frame *)malloc(sizeof(struct shm_frame));
    memset(shm_frame, 0, sizeof(struct shm_frame));

    shm_frame->frame_buf = (struct shm_frame_buf *)(shm_buf + (i * sizeof(struct shm_frame_buf))/4);
    shm_frame->frame_buf->lock = DVFRAME_READY;

    if (shm_frame_last == NULL) {
      _xdvshow_shm.shm_frame = shm_frame;
    }
    else {
      shm_frame_last->next = shm_frame;
    }
    shm_frame_last = shm_frame;
  }
  shm_frame->next = _xdvshow_shm.shm_frame;
  _xdvshow_shm.read_frame = _xdvshow_shm.shm_frame;

  shm_frame = NULL;
  shm_frame_last = NULL;
  for (i=0; i<_xdvshow_shm.framebufs; i++) {
    shm_frame = (struct shm_frame *)malloc(sizeof(struct shm_frame));
    memset(shm_frame, 0, sizeof(struct shm_frame));

    shm_frame->frame_buf = (struct shm_frame_buf *)(shm_buf + ((_xdvshow_shm.framebufs + i) * sizeof(struct shm_frame_buf))/4);
    shm_frame->frame_buf->lock = DVFRAME_READY;

    if (shm_frame_last == NULL) {
      _xdvshow_shm.audio_shm_frame = shm_frame;
    }
    else {
      shm_frame_last->next = shm_frame;
    }
    shm_frame_last = shm_frame;
  }
  shm_frame->next = _xdvshow_shm.audio_shm_frame;

  return(1);
}
